package edu.gmu.cs475;


import java.nio.file.DirectoryNotEmptyException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.List;
import java.util.Map;
import java.util.Set;

public abstract class AbstractKeyValueClient {

	protected IKeyValueServer keyValueServer;

	protected AbstractKeyValueClient(String host, int port) {
		try {
			Registry registry = LocateRegistry.getRegistry(host, port);
			keyValueServer = (IKeyValueServer) registry.lookup(IKeyValueServer.RMI_NAME);
		} catch (Exception e) {
			System.err.println("Client exception connecting to lock server: " + e.toString());
			e.printStackTrace();
		}
	}

	protected AbstractKeyValueClient(IKeyValueServer server) {
		this.keyValueServer = server;
	}

	public String get(String key) throws RemoteException {
		return keyValueServer.get(key);
	}

	public Set<String> listKeys() throws RemoteException {
		return keyValueServer.listKeys();
	}

	public void set(String key, String value) throws RemoteException {
		keyValueServer.set(key, value);
	}

	public boolean remove(String key) throws RemoteException {
		return keyValueServer.remove(key);
	}

	public Set<String> listDirectory(String directory) throws RemoteException {
		return keyValueServer.listDirectory(directory);
	}

	public void removeDirectory(String directory) throws RemoteException, DirectoryNotEmptyException {
		keyValueServer.removeDirectory(directory);
	}

	public long lockKey(String name, boolean forWrite) throws RemoteException {
		long stamp = keyValueServer.lockKey(name, forWrite);
		lockKeySuccess(name, forWrite, stamp);
		return stamp;
	}

	public void unLockKey(String name, long stamp, boolean forWrite) throws RemoteException {
		unlockKeyCalled(name, forWrite, stamp);
		keyValueServer.unLockKey(name, stamp, forWrite);
	}

	public void heartbeat(String key, long stampId, boolean isWrite) throws RemoteException, IllegalMonitorStateException, IllegalArgumentException {
		keyValueServer.heartbeat(key, stampId, isWrite);
	}

	public List getWriteLockedKeys() throws RemoteException {
		return keyValueServer.getWriteLockedKeys();
	}

	public List getReadLockedKeys() throws RemoteException {
		return keyValueServer.getReadLockedKeys();
	}

	/**
	 * Prints out all keys that have are directly nested within the given
	 * directory. Must internally synchronize to guarantee that the list of keys
	 * within a directory does not change during its call, and that each key returned
	 * does not change during its execution (using a read/write lock). You should acquire
	 * all of the locks, then read all of the keys and release the locks. Your code should not
	 * deadlock while waiting to acquire locks.
	 *
	 * Your implementation must follow exactly this scheme:
	 * 1. Acquire all locks
	 * 2. Read all values
	 * 3. Release all locks
	 * (You may not release any locks until you are finished reading)
	 *
	 * @param directory to search within
	 * @return A map of every key=value pair directly contained by the given directory
	 * @throws IllegalArgumentException if the key does not represent a directory (does not end in a /)
	 * @throws IllegalArgumentException if the key does not start with a /
	 * @throws RemoteException          if any IOException occurs in the underlying RMI operation
	 */
	public abstract Map<String, String> getAll(String directory) throws RemoteException;

	/**
	 * Sets the content for all keys that are directly contained by the given directory.
	 * Must internally synchronize to guarantee that the list of keys in the given directory does
	 * not change during its call, and that each key being set does not
	 * change during its execution (using a read/write lock)
	 * <p>
	 * Given two concurrent calls to putAll, it will be indeterminite
	 * which call happens first and which happens last. But what you can (and
	 * must) guarantee is that all keys affected will have the *same* value (and not
	 * some the result of the first, and some the result of the second). Your
	 * code should not deadlock while waiting to acquire locks.
	 *
	 * Your implementation must follow exactly this scheme:
	 *   1. Acquire all locks
	 *   2. Write all values
	 *   3. Release all locks
	 *  (You may not release any locks until you are finished writing)
	 *
	 * @param directory Directory to do updates within
	 * @param content   The content to write out to each key
	 * @throws IllegalArgumentException if the key does not represent a directory (does not end in a /)
	 * @throws IllegalArgumentException if the key does not start with a /
	 * @throws RemoteException          if any IOException occurs in the underlying RMI operation
	 */
	public abstract void putAll(String directory, String content) throws RemoteException;


	/**
	 * A callback for you to implement to let you know that a lock was
	 * successfully acquired, and that you should start sending heartbeats for
	 * it
	 *
	 * @param name
	 * @param forWrite
	 * @param stamp
	 */
	public abstract void lockKeySuccess(String name, boolean forWrite, long stamp);

	/**
	 * A callback for you to implement to let you know a lock was relinquished,
	 * and that you should stop sending heartbeats for it.
	 *
	 * @param name
	 * @param forWrite
	 * @param stamp
	 */
	public abstract void unlockKeyCalled(String name, boolean forWrite, long stamp);

}