package edu.gmu.cs475;

import org.easymock.*;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.Timeout;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

import static org.easymock.EasyMock.*;
import static org.junit.Assert.assertFalse;

public class P2HeartbeatBulkTests {

	AbstractKeyValueClient keyValueClient;

	@Rule
	public Timeout globalTimeout = new Timeout(12000);

	@Mock
	private IKeyValueServer server;

	private static final int N_FILES = 2;

	@Before
	public void setup() throws Exception {
		IMocksControl mocker = EasyMock.createControl();
		mocker.resetToStrict();
		server = mocker.createMock(IKeyValueServer.class);
		keyValueClient = new KeyValueClient(server);
		expect(server.listDirectory("/testDirectory/")).andAnswer(new IAnswer<HashSet<String>>() {

			@Override
			public HashSet<String> answer() throws Throwable {
				HashSet<String> ret = new HashSet<>();
				for (int i = 0; i < N_FILES; i++)
					ret.add("/testDirectory/keys" + i);
				return ret;
			}
		}).anyTimes();
	}

	@Test
	public void testPutAllLocksThenWritesThenUnlocks() throws Exception
	{
		expect(server.lockKey(anyString(), eq(true))).andReturn(101L).once();
		expect(server.lockKey(anyString(), eq(true))).andReturn(102L).once();
		server.set(anyString(), eq("content"));
		expectLastCall().times(N_FILES);
		server.unLockKey(anyString(), anyLong(), eq(true));
		expectLastCall().once();
		server.unLockKey(anyString(), anyLong(), eq(true));
		expectLastCall().once();

		replay(server);
		try {
			keyValueClient.putAll("/testDirectory/", "content");
		} catch (IOException ex) {
			// OK, IOException expected
		}
		verify(server);
	}

	@Test
	public void testGetAllLocksThenReadsThenUnlocks() throws Exception
	{

		expect(server.lockKey(anyString(), eq(false))).andReturn(101L).once();
		expect(server.lockKey(anyString(), eq(false))).andReturn(102L).once();
		expect(server.get(anyString())).andReturn("foo").once();
		expect(server.get(anyString())).andReturn("foo").once();
		server.unLockKey(anyString(), anyLong(), eq(false));
		expectLastCall().once();
		server.unLockKey(anyString(), anyLong(), eq(false));
		expectLastCall().once();

		replay(server);
		try {
			keyValueClient.getAll("/testDirectory/");
		} catch (IOException ex) {
			// OK, IOException expected
		}
		verify(server);
	}

	static class anyStringMatchingString implements IArgumentMatcher
	{
		private Capture<String> expected;
		int i;
		boolean called;
		public anyStringMatchingString(Capture<String> s,int i){
			this.expected = s;

		}

		private String m;
		@Override
		public boolean matches(Object argument) {
			this.m = (String)argument;
			this.called = true;
			return expected.getValue().equals(argument);
		}

		@Override
		public void appendTo(StringBuffer buffer) {
			if(called)
				buffer.append("expected "+expected.getValue() +" as locked file #"+i+", not "+m);
			else
				buffer.append("any string");

		}
	}
	private static String anyStringMatchingString(Capture<String> s,int i){
		EasyMock.reportMatcher(new anyStringMatchingString(s,i));
		return null;
	}
	@Test
	public void testGetAllLocksSameOrder() throws Exception
	{
		HashSet<String> files = new HashSet<>();
		for(int i =0;i<100;i++)
			files.add("/directory/File"+i);
		expect(server.listDirectory("/directory/")).andReturn(files).once();
		LinkedList<Capture<String>> locks = new LinkedList<>();
		for(int i = 0; i < 100; i++) {

			Capture<String> file = EasyMock.newCapture();
			expect(server.lockKey(capture(file), eq(false))).andReturn(100L + i).once();
			locks.add(file);
		}
		expect(server.get(anyString())).andReturn("foo").times(100);
		server.unLockKey(anyString(), anyLong(), eq(false));
		expectLastCall().times(100);


		List<String> files2 = new LinkedList<String>(files);
		Collections.shuffle(files2);
		expect(server.listDirectory("/directory/")).andReturn(new LinkedHashSet<>(files2)).once();
		for(int i = 0; i < 100; i++) {
			expect(server.lockKey(anyStringMatchingString(locks.get(i),i),eq(false))).andReturn(500L+i).once();
		}
		expect(server.get(anyString())).andReturn("foo").times(100);
		server.unLockKey(anyString(), anyLong(), eq(false));
		expectLastCall().times(100);

		replay(server);
		try {
			keyValueClient.getAll("/directory/");
		} catch (IOException ex) {
			// OK, IOException expected
		}
		try {
			keyValueClient.getAll("/directory/");
		} catch (IOException ex) {
			// OK, IOException expected
		}
		try {
			verify(server);
		} catch (AssertionError er) {
			String[] lines = er.getMessage().split("\n");
			for (int i = 0; i < Math.min(5, lines.length); i++) {
				System.err.println(lines[i]);
			}
			if (lines.length > 5)
				System.err.println("\t... and " + (lines.length - 5) + " more");
			Assert.fail();
		}
	}
	@Test
	public void testPutAllLocksSameOrder() throws Exception
	{
		LinkedList<String> files = new LinkedList<>();
		for(int i =0;i<100;i++)
			files.add("/directory/File"+i);
		expect(server.listDirectory("/directory/")).andReturn(new LinkedHashSet<>(files)).once();
		LinkedList<Capture<String>> locks = new LinkedList<>();
		for(int i = 0; i < 100; i++) {

			Capture<String> file = EasyMock.newCapture();
			expect(server.lockKey(capture(file), eq(true))).andReturn(100L + i).once();
			locks.add(file);
		}
		server.set(anyString(), eq("content"));
		expectLastCall().times(100);
		server.unLockKey(anyString(), anyLong(), eq(true));
		expectLastCall().times(100);


		List<String> files2 = new LinkedList<String>(files);
		Collections.shuffle(files2);
		expect(server.listDirectory("/directory/")).andReturn(new LinkedHashSet<>(files2)).once();
		for(int i = 0; i < 100; i++) {
			expect(server.lockKey(anyStringMatchingString(locks.get(i),i),eq(true))).andReturn(500L+i).once();
		}
		server.set(anyString(), eq("content2"));
		expectLastCall().times(100);
		server.unLockKey(anyString(), anyLong(), eq(true));
		expectLastCall().times(100);

		replay(server);
		try {
			keyValueClient.putAll("/directory/","content");
		} catch (RemoteException ex) {
			// OK, IOException expected
		}
		try {
			keyValueClient.putAll("/directory/","content2");
		} catch (RemoteException ex) {
			// OK, IOException expected
		}
		try {
			verify(server);
		} catch (AssertionError er) {
			String[] lines = er.getMessage().split("\n");
			for (int i = 0; i < Math.min(5, lines.length); i++) {
				System.err.println(lines[i]);
			}
			if (lines.length > 5)
				System.err.println("\t... and " + (lines.length - 5) + " more");
			Assert.fail();
		}
	}



	private volatile boolean err;
	@Test
	public void testCantWriteSameFileTwiceAtOnce() throws Exception {
		final KeyValueServer server = new KeyValueServer();
		server.set("/directory/a", "b");
		server.set("/directory/b", "b");
		server.set("/directory/c", "b");
		final long TIMEOUT = 500;
		this.err = false;
		//We can make a fake client that blocks in its write then make sure that another client can't get through at the same time
		AtomicInteger nConcurrentWrites = new AtomicInteger();
		AbstractKeyValueClient slowWriter = new KeyValueClient(server) {
			@Override
			public void set(String file, String content) throws RemoteException {
				if (nConcurrentWrites.incrementAndGet() > 1) {
					System.err.println("\nError: Detected concurrent writes to " + file);
					err = true;
					return;
				}
				try {
					Thread.sleep(TIMEOUT);
				} catch (InterruptedException ex) {
					ex.printStackTrace();
				}
				nConcurrentWrites.decrementAndGet();
			}
		};

		Thread writerThread = new Thread(() -> {
			try {

				slowWriter.putAll("/directory/", "content1");
			} catch (IOException e) {
				e.printStackTrace();
				err = true;
			}
		});

		Thread writer2Thread = new Thread(() -> {
			try {
				slowWriter.putAll("/directory/", "content2");
			} catch (IOException e) {
				e.printStackTrace();
				err = true;
			}
		});
		writerThread.start();
		writer2Thread.start();
		writerThread.join();
		writer2Thread.join();
//		server.cacheDisconnect("fake client", 90040);
		assertFalse("See error above", err);
	}

	@Test
	public void testCanWriteDifferentFilesAtOnce() throws Exception {
		final KeyValueServer server = new KeyValueServer();

		server.set("/directory/a", "b");
		server.set("/directory/b", "b");
		server.set("/directory/c", "b");

		server.set("/directorya/a", "b");
		server.set("/directorya/b", "b");
		server.set("/directorya/c", "b");

		final long TIMEOUT = 500;
		this.err = true;
		//We can make a fake client that blocks in its write then make sure that another client can't get through at the same time
		AtomicInteger nConcurrentWrites = new AtomicInteger();
		KeyValueClient slowWriter = new KeyValueClient(server) {
			@Override
			public void set(String file, String content) throws RemoteException {
				if (nConcurrentWrites.incrementAndGet() > 1) {
					err = false;
					return;
				}
				try {
					Thread.sleep(TIMEOUT);
				} catch (InterruptedException ex) {
					ex.printStackTrace();
				}
				nConcurrentWrites.decrementAndGet();
			}
		};

		Thread writerThread = new Thread(() -> {
			try {

				slowWriter.putAll("/directory/", "content1");
			} catch (IOException e) {
				e.printStackTrace();
				err = true;
			}
		});

		Thread writer2Thread = new Thread(() -> {
			try {
				slowWriter.putAll("/directorya/", "content2");
			} catch (IOException e) {
				e.printStackTrace();
				err = true;
			}
		});
		writerThread.start();
		writer2Thread.start();
		writerThread.join();
		writer2Thread.join();
		assertFalse("Expected to be able to putAll to different directories at same time", err);
	}

	@Test
	public void testCanGetAllTwiceAtOnce() throws Exception {
		final KeyValueServer server = new KeyValueServer();
		server.set("/directory/a", "b");
		server.set("/directory/b", "b");
		server.set("/directory/c", "b");

		final long TIMEOUT = 500;
		this.err = true;
		//We can make a fake client that blocks in its write then make sure that another client can't get through at the same time
		AtomicInteger nConcurrentWrites = new AtomicInteger();
		AbstractKeyValueClient slowWriter = new KeyValueClient(server) {
			@Override
			public String get(String file) throws RemoteException {
				if (nConcurrentWrites.incrementAndGet() > 1) {
					err = false;
					return "content";
				}
				try {
					Thread.sleep(TIMEOUT);
				} catch (InterruptedException ex) {
					ex.printStackTrace();
				}
				nConcurrentWrites.decrementAndGet();
				return "content";
			}
		};

		Thread writerThread = new Thread(() -> {
			try {

				slowWriter.getAll("/directory/");
			} catch (IOException e) {
				e.printStackTrace();
				err = true;
			}
		});

		Thread writer2Thread = new Thread(() -> {
			try {
				slowWriter.getAll("/directory/");
			} catch (IOException e) {
				e.printStackTrace();
				err = true;
			}
		});
		writerThread.start();
		writer2Thread.start();
		writerThread.join();
		writer2Thread.join();
		assertFalse("Expected to be able to getAll from the same directory at same time", err);
	}

}
