package edu.gmu.cs475;

import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.LinkedList;

import org.easymock.Mock;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.Timeout;

import static org.easymock.EasyMock.*;

public class P2HeartbeatClientTests {
	AbstractKeyValueClient keyValueClient;

	@Rule
	public Timeout globalTimeout = new Timeout(22000);
	
	@Mock
	private IKeyValueServer server;

	@Before
	public void setup() throws Exception {
		server = mock(IKeyValueServer.class);
		keyValueClient = new KeyValueClient(server);
	}

	/**
	 * This test will take a write lock, then wait 5 seconds. In that time,
	 * there should be 2 heartbeats. Then, it will unlock the file, and sleep
	 * another 5 seconds, in which there should be 0 heartbeats.
	 * 
	 * @throws Exception
	 */
	@Test
	public void testLockSleepHeartbeatUnlock() throws Exception {
		String key = "testLockSleepHeartbeatUnlock.file0";
		expect(server.lockKey(key, true)).andReturn(101L);
		server.heartbeat(key, 101L, true);
		expectLastCall().times(2);
		server.unLockKey(key, 101L, true);
		expectLastCall();
		replay(server);
		long stamp = keyValueClient.lockKey(key, true);
		Thread.sleep(5000);
		keyValueClient.unLockKey(key, stamp, true);
		Thread.sleep(5000);

		verify(server);
	}

	/**
	 * This test will try to lock a file twice for read, then wait 5 seconds. In
	 * that time, there should be 4 heartbeats -- two from each file. Then, it
	 * will unlock the file once, wait 1 second and expect a single heartbeat,
	 * then unlock the other.
	 * 
	 * @throws Exception
	 */
	@Test
	public void testLockSameFileTwice() throws Exception {
		String key = "testLockSameFileTwice.file0";
		expect(server.lockKey(key, false)).andReturn(101L).once();
		expect(server.lockKey(key, false)).andReturn(102L).once();
		server.heartbeat(key, 101L, false);
		expectLastCall().times(2);
		server.heartbeat(key, 102L, false);
		expectLastCall().times(3);
		server.unLockKey(key, 101L, false);
		expectLastCall().once();
		server.unLockKey(key, 102L, false);
		expectLastCall().once();
		replay(server);
		long stamp1 = keyValueClient.lockKey(key, false);
		long stamp2 = keyValueClient.lockKey(key, false);
		Thread.sleep(5000);
		keyValueClient.unLockKey(key, stamp1, false);
		Thread.sleep(1500);
		keyValueClient.unLockKey(key, stamp2, false);
		Thread.sleep(3000);

		verify(server);
	}

	/**
	 * This test will try to cat all files, but the server will throw a
	 * RemoteException on one of the reads. The test will assert that you still call
	 * unlock on all files.
	 * 
	 * @throws Exception
	 */
	@Test
	public void testGetAllExceptionsCauseUnlock() throws Exception {
		LinkedList<String> files = new LinkedList<>();
		String key1 = "/testCatAllExceptionsCauseUnlock/file0";
		files.add(key1);
		String key2 = "/testCatAllExceptionsCauseUnlock/file1";
		files.add(key2);
		expect(server.listDirectory("/testCatAllExceptionsCauseUnlock/")).andReturn(new HashSet<>(files)).once();
		expect(server.lockKey(key1, false)).andReturn(101L).once();
		expect(server.lockKey(key2, false)).andReturn(102L).once();
		expect(server.get(key1)).andReturn("File 0 is ok!").once();
		expect(server.get(key2)).andThrow(new RemoteException()).once();
		server.unLockKey(key1, 101L, false);
		expectLastCall().once();
		server.unLockKey(key2, 102L, false);
		expectLastCall().once();

		replay(server);
		try {
			keyValueClient.getAll("/testCatAllExceptionsCauseUnlock/");
		} catch (RemoteException ex) {
			// OK, RemoteException expected
		}
		verify(server);
	}

	/**
	 * This test will try to put to a directory, but the server will throw a
	 * RemoteException on one of the write. The test will assert that you still call
	 * unlock on all files.
	 * 
	 * @throws Exception
	 */
	@Test
	public void testPutAllExceptionsCauseUnlock() throws Exception {
		LinkedList<String> files = new LinkedList<>();
		String key1 = "/testEchoAllExceptionsCauseUnlock/file0";
		files.add(key1);
		String key2= "/testEchoAllExceptionsCauseUnlock/file1";
		files.add(key2);
		expect(server.listDirectory("/testEchoAllExceptionsCauseUnlock/")).andReturn(new HashSet<>(files)).once();
		expect(server.lockKey(key1, true)).andReturn(101L).once();
		expect(server.lockKey(key2, true)).andReturn(102L).once();
		server.set(anyString(), anyString());
		expectLastCall().once();
		server.set(anyString(), anyString());
		expectLastCall().andThrow(new RemoteException());
		server.unLockKey(key1, 101L, true);
		expectLastCall().once();
		server.unLockKey(key2, 102L, true);
		expectLastCall().once();

		replay(server);
		try {
			keyValueClient.putAll("/testEchoAllExceptionsCauseUnlock/", "content");
		} catch (RemoteException ex) {
			// OK, RemoteException expected
		}
		verify(server);
	}

	/**
	 * This test will lock a file for read twice, then throw an
	 * IllegalMonitorStateException from one of the heartbeats and make sure
	 * that neither heartbeat nor unlock get called on that file; the other one
	 * should still heartbeat.
	 * 
	 * @throws Exception
	 */
	@Test
	public void testExceptionInHeartbeatCancelsHeartbeat() throws Exception {
		String key = "/testExceptionInHeartbeatCancelsHeartbeat/file0";
		expect(server.lockKey(key, false)).andReturn(101L).once();
		expect(server.lockKey(key, false)).andReturn(102L).once();

		server.heartbeat(key, 101L, false);
		expectLastCall().andThrow(new IllegalMonitorStateException());
		server.heartbeat(key, 102L, false);
		expectLastCall().times(2);
		server.unLockKey(key, 102L, false);
		expectLastCall().once();
		replay(server);

		keyValueClient.lockKey(key, false);
		keyValueClient.lockKey(key, false);
		Thread.sleep(4500);
		keyValueClient.unLockKey(key, 102L, false);
		verify(server);

	}

	/**
	 * This test will lock a file for write, then throw an
	 * IllegalMonitorStateException the heartbeats and make sure that neither
	 * heartbeat nor unlock get called on that file.
	 * 
	 * @throws Exception
	 */
	@Test
	public void testExceptionInWriteHeartbeatCancelsHeartbeat() throws Exception {
		String key = "/testExceptionInWriteHeartbeatCancelsHeartbeat/file0";
		expect(server.lockKey(key, true)).andReturn(101L).once();

		server.heartbeat(key, 101L, true);
		expectLastCall().andThrow(new IllegalMonitorStateException());
		replay(server);

		keyValueClient.lockKey(key, true);
		Thread.sleep(4500);
		verify(server);

	}
}
