package edu.gmu.cs475;

import org.junit.*;
import org.junit.rules.Timeout;

import java.net.ServerSocket;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.LinkedList;

import static org.easymock.EasyMock.*;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

public class P4IntegrationTests {
	static Registry rmiRegistry;
	static int port;
	@Rule
	public Timeout globalTimeout = new Timeout(22000);
	KeyValueClient keyValueClient;
	KeyValueServer lockServer;

	/**
	 * Configure a single RMI Registry to use for all tests
	 *
	 * @throws Exception
	 */
	@BeforeClass
	public static void setupRMI() throws Exception {
		// Get a free port and set up a server, client
		try (ServerSocket socket = new ServerSocket(0)) {
			socket.setReuseAddress(true);
			port = socket.getLocalPort();
		}
		rmiRegistry = LocateRegistry.createRegistry(port);
	}

	/**
	 * Create a new server and client
	 */
	@Before
	public void setup() throws Exception {
		lockServer = new KeyValueServer();
		IKeyValueServer stub = (IKeyValueServer) UnicastRemoteObject.exportObject(lockServer, 0);
		rmiRegistry.rebind(IKeyValueServer.RMI_NAME, stub);
		keyValueClient = new KeyValueClient("localhost", port);
	}

	@After
	public void cleanup() throws Exception {
		rmiRegistry.unbind(IKeyValueServer.RMI_NAME);
	}

	private String getAKey() {
		StackTraceElement[] st = Thread.currentThread().getStackTrace();
		String testMethod = "/tests/p4/" + st[2].getMethodName();
		lockServer.set(testMethod, "testData");
		return testMethod;
	}

	/**
	 * Take out a write lock on a single key, make sure heartbeat works
	 */
	@Test
	public void testLockKeyAutoHeartbeatWorks() {
		try {
			String key = getAKey();
			long stamp = keyValueClient.lockKey(key, true);
			Thread.sleep(5000);
			keyValueClient.unLockKey(key, stamp, true);
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}
	}

	/**
	 * Take out multiple read locks, make sure they all get their heartbeats in
	 */
	@Test
	public void testLockKeyMultipleReadlocksAutoHeartbeatWorks() {
		try {
			String key = getAKey();
			long stamp = keyValueClient.lockKey(key, false);
			long stamp1 = keyValueClient.lockKey(key, false);
			long stamp2 = keyValueClient.lockKey(key, false);
			long stamp3 = keyValueClient.lockKey(key, false);
			Thread.sleep(5000);
			keyValueClient.unLockKey(key, stamp, false);
			keyValueClient.unLockKey(key, stamp1, false);
			keyValueClient.unLockKey(key, stamp2, false);
			keyValueClient.unLockKey(key, stamp3, false);
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}
	}

	/**
	 * This test will prevent your client heartbeat from sending out, and then
	 * will make sure that the client's lock is removed, and that the client can
	 * re-acquire the lock
	 */
	@Test
	public void testStallHeartbeatCanReLockLater() {
		try {
			String key = getAKey();
			AbstractKeyValueClient mockedClient = partialMockBuilder(KeyValueClient.class).withConstructor("localhost", port).addMockedMethod("heartbeat").createMock();
			mockedClient.heartbeat(eq(key), anyLong(), eq(true));
			expectLastCall().times(2, 3);
			replay(mockedClient);
			long stamp = mockedClient.lockKey(key, true);
			Thread.sleep(5000);
			verify(mockedClient);

			//Next one is expected to throw an exception
			boolean caught = false;
			try {
				mockedClient.keyValueServer.heartbeat(key, stamp, true);
			} catch (IllegalMonitorStateException ex) {
				caught = true;
			}
			assertTrue("Expected IllegalMonitorStateException", caught);

			//Unlock should throw an exception as well, since the lock was canceled by the server
			caught = false;
			try {
				mockedClient.unLockKey(key, stamp, true);
			} catch (IllegalMonitorStateException ex) {
				caught = true;
			}
			assertTrue("Expected IllegalMonitorStateException", caught);

			stamp = mockedClient.lockKey(key, true); // should now be able to
			// re-acquire
			mockedClient.unLockKey(key, stamp, true);
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}
	}

	/**
	 * This test will create multiple clients, all read locking the same key.
	 * Some will be mocked to prevent their heartbeat from being delivered. We
	 * assert that the ones that succeeded in their heartbeats can unlock and
	 * the rest can't.
	 */
	@Test
	public void testMultipleClientsLockSamekeySomeFailingToHeartbeat() {
		try {
			String key = getAKey();

			AbstractKeyValueClient[] clients = new AbstractKeyValueClient[5];
			for (int i = 0; i < clients.length; i++)
				clients[i] = new KeyValueClient("localhost", port);
			AbstractKeyValueClient[] mockedClients = new AbstractKeyValueClient[5];
			for (int i = 0; i < mockedClients.length; i++) {
				AbstractKeyValueClient mockedClient = partialMockBuilder(KeyValueClient.class).withConstructor("localhost", port).addMockedMethod("heartbeat").createMock();
				mockedClient.heartbeat(eq(key), anyLong(), eq(true));
				expectLastCall().times(2, 3);
				replay(mockedClient);
				mockedClients[i] = mockedClient;
			}
			long[] stamps = new long[clients.length];
			long[] badStamps = new long[mockedClients.length];

			for (int i = 0; i < clients.length; i++) {
				stamps[i] = clients[i].lockKey(key, false);
			}
			for (int i = 0; i < mockedClients.length; i++) {
				badStamps[i] = mockedClients[i].lockKey(key, false);
			}
			Thread.sleep(5000);
			for (int i = 0; i < clients.length; i++) {
				clients[i].unLockKey(key, stamps[i], false);
			}
			for (int i = 0; i < mockedClients.length; i++) {
				boolean caught = false;
				try {
					mockedClients[i].unLockKey(key, badStamps[i], false);
				} catch (IllegalMonitorStateException ex) {
					caught = true;
				}
				assertTrue("IllegalMonitorStateException expected from clients who didn't send heartbeat", caught);
			}
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}
	}

	/**
	 * This test will create multiple clients, all read locking the same key.
	 * Then, they will all try to unlock.
	 */
	@Test
	public void testMultipleClientsLockSamekey() {
		try {
			String key = getAKey();

			AbstractKeyValueClient[] clients = new AbstractKeyValueClient[10];
			for (int i = 0; i < clients.length; i++)
				clients[i] = new KeyValueClient("localhost", port);
			LinkedList<Long> stamps = new LinkedList<>();
			for (AbstractKeyValueClient c : clients) {
				stamps.add(c.lockKey(key, false));
			}
			Thread.sleep(5000);
			for (AbstractKeyValueClient c : clients) {
				c.unLockKey(key, stamps.pop(), false);
			}
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}
	}

}
