package edu.gmu.cs475;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.NoSuchFileException;
import java.rmi.RemoteException;
import java.util.stream.Collectors;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.Timeout;

import edu.gmu.cs475.internal.ServerMain;

public class P3ServerTests {
	KeyValueServer server;
	@Rule
	public Timeout globalTimeout = new Timeout(22000);
	
	@Before
	public void setup() throws IOException {
		server = new KeyValueServer();
	}

	/**
	 * This test will see that a lock expires after the stated time of no
	 * heartbeats
	 * 
	 */
	@Test
	public void testLockExpires() throws RemoteException {
		try {
			String key = "/tests/testLockExpires";
			server.set(key,"baz");
			long stamp = server.lockKey(key, true);
			Thread.sleep(4000);
			boolean caught = false;
			try {
				server.unLockKey(key, stamp, true);
			} catch (IllegalMonitorStateException ex) {
				/* Expected */
				caught = true;
			}
			assertTrue("IllegalMonitorStateException should be thrown after lock times out", caught);
			assertTrue("There should be no active locks", server.getWriteLockedKeys().isEmpty());
			assertTrue("There should be no active locks", server.getReadLockedKeys().isEmpty());
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}

	}

	/**
	 * This test will try to keep a lock alive by sending heartbeats at the
	 * appropriate interval
	 * 
	 */
	@Test
	public void testLockHeartbeatsRenewLock() throws RemoteException {
		try {
			String key = "/foo/testLockHeartbeatsRenewLock";
			server.set(key,"baz");
			long stamp = server.lockKey(key, true);
			Thread.sleep(2000);
			server.heartbeat(key, stamp, true);
			Thread.sleep(2000);
			server.heartbeat(key, stamp, true);
			Thread.sleep(2000);
			server.unLockKey(key, stamp, true);
			assertTrue("There should be no active locks", server.getWriteLockedKeys().isEmpty());
			assertTrue("There should be no active locks", server.getReadLockedKeys().isEmpty());
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}
	}

	/**
	 * This test will send heartbeats at an irregular interval (every 1 second),
	 * testing that each heartbeat renews the countdown timer to expire the lock
	 * by the full amount
	 * 
	 */
	@Test
	public void testIrregularHeartbeatStillRenewLock() throws RemoteException  {
		try {
			String key = "/foo/testIrregularHeartbeatStillRenewLock";
			server.set(key,"baz");

			long stamp = server.lockKey(key, true);
			Thread.sleep(1000);
			server.heartbeat(key, stamp, true);
			Thread.sleep(1000);
			server.heartbeat(key, stamp, true);
			Thread.sleep(2500);
			server.unLockKey(key, stamp, true);
			assertTrue("There should be no active locks", server.getWriteLockedKeys().isEmpty());
			assertTrue("There should be no active locks", server.getReadLockedKeys().isEmpty());
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}
	}

	/**
	 * This test will take out several read locks on the same file and let some
	 * expire. It will check that the correct lock got canceled.
	 * 
	 */
	@Test
	public void testHeartbeatOnlyExpiresCorrectLock() throws RemoteException {
		try {
			String key = "/foo/testHeartbeatOnlyExpiresCorrectLock";
			server.set(key,"baz");

			long stamp1 = server.lockKey(key, false);
			long stamp2 = server.lockKey(key, false);
			long stamp3 = server.lockKey(key, false);
			long stamp4 = server.lockKey(key, false);
			Thread.sleep(2000);
			server.heartbeat(key, stamp2, false);
			server.heartbeat(key, stamp3, false);
			server.heartbeat(key, stamp4, false);
			Thread.sleep(2000);
			server.heartbeat(key, stamp2, false);
			server.heartbeat(key, stamp3, false);
			server.heartbeat(key, stamp4, false);
			server.unLockKey(key, stamp2, false);
			boolean caught = false;
			try {
				server.unLockKey(key, stamp1, false);
			} catch (IllegalMonitorStateException ex) {
				// expected
				caught = true;
			}
			server.unLockKey(key, stamp3, false);
			server.unLockKey(key, stamp4, false);
			assertTrue("stamp1 should have become invalid", caught);
		} catch (IllegalMonitorStateException | IllegalArgumentException | InterruptedException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		}
	}

	/**
	 * This test will try to send a variety of invalid heartbeats, making sure
	 * they throw the appropriate exceptions
	 * 
	 * @throws Exception
	 */
	@Test
	public void testHeartbeatInvalidStamp() throws RemoteException {
		String key = "/foo/testHeartbeatInvalidStamp";
		server.set(key,"baz");
		long stamp1 = -1;
		try {
			stamp1 = server.lockKey(key, false);
			boolean caught = false;
			try {
				server.heartbeat(key, stamp1 + 1, false);
			} catch (IllegalMonitorStateException ex) {
				caught = true;
			}
			assertTrue("Expected illegal monitor state exception to be thrown", caught);
			try {
				server.heartbeat(key, stamp1, true);
			} catch (IllegalMonitorStateException ex) {
				caught = true;
			}
			assertTrue("Expected illegal monitor state exception to be thrown", caught);
			try {
				server.heartbeat(key + ".fake", stamp1, false);
			} catch (IllegalArgumentException ex) {
				caught = true;
			}
			assertTrue("Expected NoSuchFileException to be thrown", caught);
		} catch (IllegalMonitorStateException | IllegalArgumentException | RemoteException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		} finally {
			if (stamp1 >= 0)
				try {
					server.unLockKey(key, stamp1, false);
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
					fail("Unexpected exception");
				}
		}
	}

	/**
	 * This test will try to send a variety of invalid unlocks, making sure
	 * they throw the appropriate exceptions
	 * 
	 * @throws Exception
	 */
	@Test
	public void testUnlockInvalidStamp() throws RemoteException {
		String key = "/foo/testUnlockInvalidStamp";
		server.set(key,"baz");
		long stamp1 = -1;
		try {
			stamp1 = server.lockKey(key, false);
			boolean caught = false;
			try {
				server.unLockKey(key, stamp1 + 1, false);
			} catch (IllegalMonitorStateException ex) {
				caught = true;
			}
			assertTrue("Expected illegal monitor state exception to be thrown", caught);
			try {
				server.unLockKey(key, stamp1, true);
			} catch (IllegalMonitorStateException ex) {
				caught = true;
			}
			assertTrue("Expected illegal monitor state exception to be thrown", caught);
			try {
				server.unLockKey(key + ".fake", stamp1, false);
			} catch (IllegalArgumentException ex) {
				caught = true;
			}
			assertTrue("Expected IllegalArgumentException to be thrown", caught);
		} catch (IllegalMonitorStateException | IllegalArgumentException ex) {
			ex.printStackTrace();
			fail("Unexpected exception");
		} finally {
			if (stamp1 >= 0)
				try {
					server.unLockKey(key, stamp1, false);
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
					fail("Unexpected exception");
				}
		}
	}
}
